package math;
//package scoot.math.linalg;

/**
  An LU decomposition of a matrix which can be used to back-substitute
  to solve Ax = b for x.  
  The same LUDState may be used in multiple backsubs of the same matrix.
  This class is <B>IMMUTABLE</B>.
  */

public class LUDState
{
  /*************
    Rep
    ************/
  protected Matrix lud;
  protected int[] permutation;
  protected int d;
  protected Vec scratchColumn;
  
  /** non public constructor.  Matrix class only constructs this
     and returns it. */
  protected LUDState(Matrix m)
  {
    lud = new Matrix(m);
    this.permutation = new int[m.numRows()];
    scratchColumn = new Vec(m.numRows());
  }
  
  /**
    Back substitutes to solve for x in Ax=b. 
    b is unchanged. 
    May be called on different b.
    
    @param b vector
    @return x Creates new Vec x to return. 
    */
  public Vec backSub(Vec b)
  {
    Vec x = new Vec(b.dim());
    backSub(b, x);
    return x;
  }

  /**
    Perform the backsubstitution to solve Ax = b for x.
    Given b, returns in x.
    <B>SAFE:</B> x can be b.
    
    @param b vector
    @param x vector to store result.  Must match in dimension.
    @exception DimensionMismatchException when x.dim() != b.dim()
    
    */
  public void backSub(Vec b, Vec x) 
  {
    /* if same, no need! */
    if (x != b)
      x.copyFrom(b);
    
    int i,ii=-1,ip,j;
    double sum;
    int n = lud.numRows();
    
    for (i=0;i<n;i++) 
      {
	ip=permutation[i];
	sum = x.rep[ip];
	x.rep[ip] = x.rep[i];
	if (ii > -1)
	  {
	    // aries shoudl the j<=i-1 be <?
	    for (j=ii;j<=i-1;j++) 
	      sum -= lud.rep[i][j] * x.rep[j];
	  }
	else if (sum != 0.0) 
	  { 
	    ii=i;
	  }
	
	x.rep[i]=sum;
      }
    
    for (i = n-1; i>=0; i--) 
      {
	sum=x.rep[i];
	for (j=i+1;j<n;j++) 
	  sum -= lud.rep[i][j] * x.rep[j];
	x.rep[i] = sum / lud.rep[i][i];
      }
    
  }

  
  /**
    Calculate the determinant from the LU decomposition.
    */
  public double determinant()
  {
    double out = d;
    for (int j = 0; j < lud.numRows(); j++)
      out *= lud.ref(j,j);
    return out;
  }


  /********************
    Plastic wrap below
    ********************/
  protected static void error(String s)
  {
    System.err.println("(LUDState): " + s);
  }

	
	
}









