package math;
//package scoot.math.linalg;

import java.io.Serializable;

/** 
  A vector of arbitrary length, but once created is fixed in size.
  
  Indices are 0-based.
  
  Some ops (noted) can cause the DimensionMismatchException, which is
  a runtime exception.  If you check dims before doing adds and muls,
  you are safe to not handle these exceptions.
  */

public final class Vec implements Cloneable, Serializable
{
  /************
    Vars
    ***********/
  protected double[] rep;
  protected int d;

  /*************
    Constructors
    **************/

  public Vec() {}

  public double[] getValues() { return rep; }
  
  public void setValues(double[] a) 
  { 
    rep = a;
    d = rep.length;
  }
  
  /** create a vector of dimension dim */
  public Vec(int dim)
  {
    rep = new double[dim];
    d = rep.length;
  }

  /** construct from another Vec */
  public Vec(Vec v)
  {
    rep = copyArray(v.rep);
    d = rep.length;
  }

  /** construct from double array, creates own storage */
  public Vec(double[] v) 
  { 
    rep = copyArray(v);
    d = rep.length; 
  } 
  
  /** construct from double array, creates own storage */
  public Vec(float[] v) 
  { 
    rep = copyArray(v);
    d = rep.length; 
  } 

  /** Create a copy of the Vec */
  public Vec copy() 
  {
    return (Vec)clone(); 
  }

  /** copy values _from_ into _this_.  
    @exception DimensionMismatchException if this.dim() != from.dim()
    */
  public void copyFrom(Vec from)
  {
    if (d != from.d)
      {
	throw new DimensionMismatchException("copyFrom: different size arrays");
      }
    int i;
    for(i=0; i<d; i++)
      {
	rep[i] = from.rep[i];
      }
    d = rep.length;
  }

  public void copyFrom(Vec v, int[] ii)
  {
    int n = ii.length;
    for (int i = 0; i < n; i++)
      rep[i] = v.rep[ii[i]];
  }

  /** clone the object */
  public Object clone()
  {
    try
      {
	Vec v = (Vec)super.clone();
	v.rep = copyArray(rep);
	v.d = rep.length;
	return v;
      }
    catch (Exception e)
      {
	error(e.toString());
	return null;
      }
  }
  
  /********
    Conversion 
    *******/

  /**
    Convert this Vec into a 1 x n matrix.
    */
  public Matrix toRowVector()
  {
    Matrix a = new Matrix(1, d);
    int i;
    for (i=0; i<d; i++)
      {
	a.set(0, i, ref(i));
      }
    return a;
  }

  /**
    Convert this Vec into a n x 1  matrix.
    */
  public Matrix toColumnVector()
  {
    Matrix a = new Matrix(d, 1);
    int i;
    for (i=0; i<d; i++)
      {
	a.set(i, 0, ref(i));
      }
    return a;
  }
  
  
  /*************
    Observers
    ************/

  /** get the ith element */
  public double ref(int i)
  {
    return rep[i];
  }

  /** Return a COPY of array, cannot sideffect this instance */
  public double[] toArray()
  {
    return copyArray(rep);
  }

  /** same as ref(int) */
  public double get(int i)
  {
    return ref(i);
  }

  /** same as ref(int) */
  public double elementAt(int i)
  {
    return ref(i);
  }

  public int dimension()
  {
    return d;
  }
  
  public int dim()
  {
    return d;
  }

  public int size()
  {
    return d;
  }
  
  /** return Euclidean magnitude squared */
  public double  magSquared()
  {
    double sum = 0.0;
    for (int i=0; i<dimension(); i++)
      {
	sum += rep[i]*rep[i];
      }
    return sum;
  }

  /** return magnitude. */

  public double magnitude()
  {
    return Math.sqrt(magSquared());
  }
  


  /** return sum of components (1-norm if you like) */
  public double sum()
  {
    double sum = 0.0;
    for (int i=0; i<dimension(); i++)
      sum += rep[i];
    return sum;
  }

  /** Return a new vector = component-wise scale _this_ by _v_ 
   @exception DimensionMismatchException if different dimensions
   */
  public Vec scale(Vec v)
  {
    if (d != v.d)
      {
	throw new DimensionMismatchException("scale: different size arrays");
      }
    Vec out = new Vec(this.d);

    for (int i = 0; i < dimension(); i++)
      out.rep[i] = rep[i] * v.rep[i];
    return out;
  }


  /** Component-wise scale _a_ and _b_ and leave the result in _this_
   @exception DimensionMismatchException if different dimensions
   */
  public void scale(Vec a, Vec b)
  {
    if ((d != a.d) || (d != b.d))
      {
	throw new DimensionMismatchException("scale: different size arrays");
      }
    for (int i = 0; i < dimension(); i++)
      rep[i] = a.rep[i]*b.rep[i];
  }
  

  

  /** Returns a normalized copy of this vector. */
  public Vec direction()
  {
    Vec vdir = this.copy();
    vdir.normalize();
    
    return vdir;
  }

  /** return Euclidean magnitude of vector. */
  public double mag()
  {
    return Math.sqrt(magSquared());
  }
 
  /**
     Return the L^infinity norm, the maximum value in the vector. 
  */
  public double infinityNorm()
  {
    return max();
  }

  /**
     Return the size of the largest element or zero if size() == 0.
  */
  public double max()
  {
    if (rep.length == 0)
      return 0.0;
    
    if (rep.length == 1)
      return rep[0];
    
    double m = rep[0];
    int i;
    for (i=1; i<rep.length; i++)
    {
      if (rep[i] > m)
	m = rep[i];
    }
    return m;
  }

  /**
     Return the size of the smallest element.
     Returns zero if size() == 0.
  */
  public double min()
  {
    if (rep.length == 0)
      return 0.0;
    
    if (rep.length == 1)
      return rep[0];
    
    double m = rep[0];
    int i;
    for (i=1; i<rep.length; i++)
    {
      if (rep[i] < m)
	m = rep[i];
    }
    return m;
  }
  
  /** dot product _this_ dot _v_ 
   @exception DimensionMismatchException if different dimensions
   */
  public double dot(Vec v)
  {
    if (d != v.d)
      throw new DimensionMismatchException("can't dot differing dimension vectors");

    double sum = 0.0;
    for(int i=0; i<dimension(); i++)
      sum += rep[i] * v.rep[i];
    return sum;
  }
  
  /** 
    angle between _this_ and _b_ in radians 
    @exception DimensionMismatchException if different dimension vectors
    */
  public double angleBetween(Vec b) 
  {
    double amag = this.mag();
    double bmag = b.mag();
    if (amag == 0.0 || bmag == 0.0)
      return 0.0;
    
    double d = this.dot(b);
    d /= (amag * bmag);
    //return scoot.math.ScootMath.acos(d);
    return Math.acos(d);
  }

  /** 
    cross product _this_ and _b_.
    @exception IllegalArgumentException if this or b is not of dimension 3.
    */
  public Vec cross(Vec b) throws IllegalArgumentException
  {
    if (d != 3 ||
	b.d != 3)
    {
      throw new IllegalArgumentException("this.cross(Vec b, Vec out): this and b must dimension 3 vectors");
    }
	
    Vec cross = new Vec(3);
    cross.rep[0] =  rep[1]*b.rep[2] - rep[2]*b.rep[1];
    cross.rep[1] = -rep[0]*b.rep[2] + rep[2]*b.rep[0];
    cross.rep[2] =  rep[0]*b.rep[1] - rep[1]*b.rep[0];
    return cross;
  }

  /** 
    cross product _this_ and _b_, store the result in _out_. It is acceptable
    for this vector to be either of the two arguments.
    @exception IllegalArgumentException if this or b is not of dimension 3.
    */
  public void cross(Vec b, Vec out) throws IllegalArgumentException
  {
    if (d != 3 ||
	b.d != 3)
    {
      throw new IllegalArgumentException("this.cross(Vec b): this and b must dimension 3 vectors");
    }
    
    // in case output is larger than input
    if (out.d > 3)
      out.zero();
    
    double x =  rep[1]*b.rep[2] - rep[2]*b.rep[1];
    double y = -rep[0]*b.rep[2] + rep[2]*b.rep[0];
    double z =  rep[0]*b.rep[1] - rep[1]*b.rep[0];

    out.rep[0] = x;
    out.rep[1] = y;
    out.rep[2] = z;
  }


  /** 
    outer product of _this_ and _b_.
    */
  public Matrix outer(Vec b)
  {
    int xn = dimension();
    int yn = b.dimension();
    Matrix A = new Matrix(xn,yn);
    for (int i = 0; i < xn; i++)
      for (int j = 0; j < yn; j++)
	A.set(i,j, ref(i)*b.ref(j));
    return A;
  }


  /** 
    outer product of _this_ and _b_, store the result in _out_.
    */
  public void outer(Vec b, Matrix out)
  {
    int xn = dimension();
    int yn = b.dimension();
    for (int i = 0; i < xn; i++)
      for (int j = 0; j < yn; j++)
	out.set(i,j, ref(i)*b.ref(j));
  }


  public String toString()
  {
    if (d == 0) return "empty";
    String s = "";
    for(int i=0; i<d-1; i++)
      {
	s = s + Format.format(rep[i], true) + " ";
      }
    s = s + Format.format(rep[d-1], true);
    return s;
  }
  
  /*****************
    Mutators
    ****************/
  
  /** resize the Vec to be the new size.  Truncates or zero pads as needed. 
   If the size is already correct, nothing happens. */
  public void resize(int n)
  {
    if (rep.length == n) return;
    
    double[] new_rep = new double[n];
    int i;
    for  (i=0; i<n; i++)
    {
      if (i < rep.length) 
	new_rep[i] = rep[i]; 
      else
	new_rep[i] = 0.0;
    }
    rep = new_rep;
    d = rep.length;
  }


  /** set me to be copy of my argument */
  public void fromVec(Vec v)
  {
    rep = copyArray(v.rep);
    d = rep.length;
  }

  public void setValue(Vec v)
  {
    fromVec(v);
  }
  
  public void setValue(float[] v)
  {
    int i; 
    resize(v.length);
    for (i=0; i<v.length; i++)
      set(i, v[i]);
  }
  
  /** set me to be copy of my argument */
  public void set(Vec v)
  {
    fromVec(v);
  }

  /** set the ith element to _val_ */
  public void set(int i, double val)
  {
    rep[i] = val;
  }

  /** side-effect the vector by scaling all elements by scalar _d_ */
  public void scale(double d)
  {
    for (int i=0; i<dimension(); i++)
      rep[i] *= d;
  }

  public void add(int i, double x)
  {
    rep[i] += x;
  }

  public void sub(int i, double x)
  {
    rep[i] -= x;
  }

  /** side-effect the vector by raising all elements by power _p_ */
  public void pow(double p)
  {
    for (int i=0; i<dimension(); i++)
      rep[i] = Math.pow(rep[i],p);
  }


  /** side-effect the vector to the zero vector */
  public void zero()
  {
    for (int i=0; i<dimension(); i++)
      rep[i] = 0;
  }
  
  /** side-effect the vector so it has unit magnitude */
  public void normalize()
  {
    double d = mag();
    if (d != 0.0)
      scale(1.0/d);
  }

  /** 
    Return a new vector _v_ =  _this_ + _b_ 
   @exception DimensionMismatchException for mismatch dimensions
   */
  public Vec add(Vec b)
  {
    Vec out = new Vec(this.d);
    add(this, b, out);
    return out;
  }
 
  /** 
    Return a new vector _v_ =  _this_ - _b_ 
    @exception DimensionMismatchException for mismatched dimension
   */
  public Vec sub(Vec b)
  {
    Vec out = new Vec(this.d);
    sub(this, b, out);
    return out;
  }
  
  
  /*******************
    Static cling, baby
    ******************/
  
  /**
    Adds a and b and stores into out.  a, b, and out MUST be the same 
    dimension or out will not be valid.
    @exception DimensionMismatchException if dimensions are wrong
    It IS safe for a == out or b == out.
    */
  public static void add(Vec a, Vec b, Vec out)
  {
    if (a.d != b.d ||
	a.d != out.d)
      {
	throw new DimensionMismatchException("add: can't add Vec with different dimension");
      }
    for(int i=0; i < a.d; i++)
      {
	out.rep[i] = a.rep[i] + b.rep[i];
      }
  }


  public static void scaleAdd(Vec a, Vec b, double scaleb, Vec out)
  {
    for (int i=0; i < a.d; i++)
      out.rep[i] = a.rep[i] + scaleb*b.rep[i];
  }


  public final void scaleAdd(double scale, Vec a)
  {
    for (int i=0; i < dimension(); i++)
      rep[i]+= scale*a.rep[i];
  }


  /**
     Subtracts a - b and stores into out.  a, b, and out MUST be the same 
     dimension or out will not be valid.
     @exception DimensionMismatchException if dimensions are wrong
     It IS safe for a == out or b == out.
     */
  public static void sub(Vec a, Vec b, Vec out)
  {
    if (a.d != b.d ||
	a.d != out.d)
      {
	throw new DimensionMismatchException("sub: can't add Vec with different dimension");
      }
    for(int i=0; i < a.d; i++)
      {
	out.rep[i] = a.rep[i] - b.rep[i];
      }
  }
  
    

  /***********************
    Dental Dams below! (thanks to Nelson)
    *********************/

  /** convenience to print to std err */
  protected static void error(String s)
  {
    System.err.println("(Vec): " + s);
  }

  protected static double[] copyArray(double[] a)
  {
    try 
      {
	return (double[])a.clone();
      }
    catch (Exception e)
      {
	error(e.toString());
	return null;
      }
  }

  protected static double[] copyArray(float[] a)
  {
    try 
      {
	double[] cpy = new double[a.length];
	for (int i = 0; i < a.length; i++)
	{
	  cpy[i] = a[i];
	}
	return cpy;
	  
      }
    catch (Exception e)
      {
	error(e.toString());
	return null;
      }
  }
  
  /** for fast multiplies within package, bare the actual rep */
  protected double[] getRep()
  {
    return rep;
  }

  protected static void print(String s)
  {
    System.out.println(s);
  }

  /************************
    Test Driver
    **************************/

  /** test driver */
  public static void main(String[] argv)
  {
    Vec a = new Vec(3);
    double[] arr = new double[3];
    arr[0] = 3.0;
    arr[1] = 0.0;
    arr[2] = 0.0;
    Vec b = new Vec(arr);
    a.set(0, 0.0);
    a.set(1, 2.0);
    a.set(2, 0.0);
    print("a: " + a);
    print("b: " + b);
    
    print("a dot b: " + a.dot(b));
    print("angle: " + a.angleBetween(b));
    print("cross: " + a.cross(b));
    print("amag2: " + a.magSquared());
    print("amag: " + a.mag());
    a.normalize();
    print("normalized a: " + a);
    print(" a dot (a cross b) better be 0.0: " + a.dot(a.cross(b)));	  

    print("a:" + a);
    print("b:" + b);
    
    print("a + b: " + a.add(b));

    print("a:" + a);
    print("b:" + b);

    print("a - b: " + a.sub(b));
    
    print("a - a: " + a.sub(a));
    
  }
}
